#
# AssocEnd.tcl
# 
# AssociationEnd code generation
#

package provide AssociationEnd 0.1;

class AssociationEnd {

  extends Element;

  import Console;
  import Comment;
  import Indent;
  import Output;
  import Expander;
  import String;
  import library;

  staticarray template;

  source [file join [library] TmplAEnd.tcl];

  field oend;
  field assoc;
  field class;

  method init {} {
    catch {
      set oend [$self get_opposite_end];
    } 
    set assoc [$self get_association];
    set class [$self get_class];
  }

  method endname {} {
    set name [$oend get_name];
    if { $name != "" } { return $name; }
    set name [$assoc get_name];
    if { $name != "" } { return $name; }
    return "[String downcase [[$oend get_class] get_name]]";
  }   

  #
  # Looks up a template that closest matches the given name
  #
  method lookup_template {name} {
    #
    # First try exact match
    #
    if { [array names template $name] == $name } {
      return $name;
    }

    #
    # Loop over all templates, looking up all that match
    #
    set tmpls {};
    foreach tmpl [array names template] {
      if { [string match $tmpl $name] } {
	lappend tmpls $tmpl;
      }
    }

    # 
    # Check what we found
    #
    switch [llength $tmpls] {
      0 { return ""; }
      1 { return $tmpls; }
    }

    #
    # Oh well, there are more than one.
    # Must select the one that matches most closely
    #
    # The algorithm is as follows:
    #  1. Calulate the number of stars in each matching template name
    #  2. Select the one with the minumun number of starts
    #  3. If there are several with an equal number of stars,
    #     select the one where the first start is latest in string
    #
    set minstars 99;
    set n 0;
    set ftmpl {};
    foreach tmpl $tmpls {
      set l [split $tmpl -];
      set ns 0;
      foreach e $l {
	if { $e == "*" } {
	  incr ns;
	}
      }
      if { $ns <= $minstars } {
	set minstars $ns;
	incr n;
	lappend ftmpl $tmpl;
	continue;
      }
    }
    if { $n == 0 } {
      return -code error "Internal error";
    }
    if { $n == 1 } {
      return $ftmpl;
    }
    #
    # Oh well, still more than one
    # If there are STILL MORE THAN one, let it be random...
    #
    set minstarindex 99;
    set stmpl {};
    foreach tmpl $ftmpl {
      set si [string first "*" $tmp];
      if { $si < $minstartindex } {
	set minstartindex $si;
	set stmpl $tmpl;
      }
    }
    #
    # Now return the one we got
    #
    return $stmpl;
  }

  #
  # Returns minimum/maximum bound given the bounds string.
  #
  staticmethod mult {type mult} {
    set min $mult;
    set max $mult;
    set l [split $mult ".."];
    if { [llength $l] > 1 } {
      set min [lindex $l 0];
      set max [lindex $l 2];
    }
    switch $type {
      max { return $max; }
      min { return $min; }
    }
    return "";
  }
    
  #
  # Generates code
  #
  method print {} {

    # "Inner class" association ends are printed by a separate method
    if { [$self is innerclass] } { 
      return;
    }

    # "Generate operations" setting is treated as "generate association end"
    if { [$oend get_gen_ops] == 0 } {
      return;
    }

    #
    # Code generation for non-binary associations not available yet
    #
    if { $oend == "" || [$assoc get_num_ends] != 2 } {
      Console warning "Code generation for higher order associations\
	  not supported";
      Console message "Not generating code for association end \
	  \"[$self endname]\" of class \"[[$self get_class] get_name]\"";
      return;
    }
    
    ###############################
    # Qualifying association attributes
    ###############################

    #
    # Qualifier
    #
    # XXX set qualifier [$self get_qualifier];

    # 
    # Multiplicity type
    #
    switch [$self get_multiplicity_type] {
      "ExactNumberGreaterThanOne" { set source_mult "Bounded"; }
      "unspecified"               { set source_mult "Unbounded"; }
      default   { set source_mult [$self get_multiplicity_type]; }
    }
    switch [$oend get_multiplicity_type] {
      "ExactNumberGreaterThanOne" { set target_mult "Bounded"; }
      "unspecified"               { set target_mult "Unbounded"; }
      default   { set target_mult [$oend get_multiplicity_type]; }
    }

    #
    # Aggregation
    #
    set aggregation [$oend get_aggregation];

    #
    # Changeability
    #
    set changeable [$oend get_changeable];

    #
    # Ordering
    #
    set isOrdered [$oend get_is_ordered];
    
    # 
    # Navigability
    #
    set navigability ""
    switch [expr [$self get_is_navigable] + 2 * [$oend get_is_navigable]] {
      3 { set navigability "<=>" }
      1 { set navigability "<=="  }
      2 { set navigability "==>"  }
    }

    # 
    # ClassDesigner specific: Access
    #
    set access [$oend get_access];

    ################################################################
    # Expander creation
    ################################################################a

    set expander [new Expander];

    #
    # Now generate the qualifying template names
    # and look up the corresponding templates. 
    # First try to use an "implement as" template name. Upon failure,
    # construct a template name from the association params.
    #
    # The template name structure is the following:
    # type-targetMultiplicity-aggregation
    #

    set tmpl_names {};
    set tmpls {};

    #
    # Generate an attribute for the association end only if
    # the corresponding class doesn't have an attribute with the same name
    #
    set gen_attr 1;
    set attributes [$class get_attributes];
    foreach attribute $attributes {
      if { [$attribute get_name] == [$self endname] } {
	Console warning "Association variable name \"[$attribute get_name]\" conflicts with another field";
	set gen_attr 0;
      }
    }
    if { $gen_attr } {
      lappend tmpl_names "attr-${target_mult}-${aggregation}";
    }

    #
    # The "Access" setting states which methods are generated 
    # for the association end
    #
    set read 1;
    set write 1;
    switch "$access" {
      readonly  { set write 0; }
      writeonly { set read 0;  }
    }

    set tmpl_name [$self get_implement_as];
    if { $tmpl_name == "" } {
      set tmpl_name "${target_mult}-${aggregation}";
    }
    lappend tmpl_names "init-$tmpl_name"
    if { $read } {
      lappend tmpl_names "get-$tmpl_name";
    }
    if { $write && $changeable == "none" } {
      lappend tmpl_names "set-$tmpl_name";
    }
    if { $write && $changeable != "frozen" && \
	($target_mult == "Bounded" || $target_mult == "Unbounded") } {
      lappend tmpl_names "add-$tmpl_name";
    }
    if { $write && $changeable != "frozen" && $navigability == "<=>" } {
      lappend tmpl_names "update-$tmpl_name";
    }
 
    foreach tmpl $tmpl_names {
      set found_tmpl [$self lookup_template $tmpl];
      if { $found_tmpl == "" } {
	Console warning "No matching template found for \
	    \"$tmpl\"";
	Console message "Not generating code for association end \
	    \"[$self endname]\" of class \"[[$self get_class] get_name]\".";
	return;
      }
      lappend tmpls $found_tmpl;
    }
    
    #
    # First create access to qualifying attributes
    #
    $expander bind isOrdered    $isOrdered;

    # 
    # Now access to REAL bounds, not multiplicity bound type
    #
    $expander bind multmin [mult min [$oend get_multiplicity]];
    $expander bind multmax [mult max [$oend get_multiplicity]];
    $expander bind multincr   10;            # XXX this value should be a property
    $expander bind lowerbound "[String upcase [$self endname]]_LOWERBOUND";
    $expander bind upperbound "[String upcase [$self endname]]_UPPERBOUND";
    $expander bind increment  "[String upcase [$self endname]]_INCREMENT";

    #
    # Bind generic code generation issues
    #
    $expander bind static     [$oend prop "static"];
    $expander bind visibility [$oend prop "visibility"];
    if { [$oend is static] } {
      $expander bind type     "Class ";# XXX (Java specific)
    } else {
      $expander bind type     [[$oend get_class] get_name];
    }
    $expander bind name       [String downcase [$self endname]];
    $expander bind Name       [String capitalize [$self endname]];
    # XXX $expander bind qualifier  [$qualifier get_name]; 

    #
    # Bind association name for comments
    # 
    set aname "[$assoc get_name]";
    if { $aname != {} } {
      $expander bind association $aname;
    } else {
      $expander bind association "Unnamed association";
    }
    
    Output print [$expander expand $template(begin)];

    #
    # Generate a comment for the association
    #
    Comment print description "[$assoc get_description]";
    Output newline;

    foreach tmpl $tmpls {
      Output print [$expander expand $template($tmpl)];
    }

    Output print [$expander expand $template(end)];

    delete $expander;
  }

  method print_as_innerclass {} {

    if { $oend == "" || [$assoc get_num_ends] != 2 } {
      Console warning "Total association multiplicity for inner classes \
	  must be 2";
      Console message "Not generating code for association end \
	  \"[$self endname]\" of class \"[[$self get_class] get_name]\"";
      return;
    }

    set class [$self get_class];
    $class print;
  }

}
