#
# Componen.tcl
#
# Routines to handle ClassDesigner components
#


package provide Component 0.1;

class Component {

  extends Element;

  import sc_isFileWritable;
  import sc_copyResourceFork;
  import sc_copyCreatorAndType;
  import sc_setCreatorAndType;
  import sc_deleteFile;
  import sc_renameFile;
  import sc_platform;
 
  import Console;
  import Output;
  import Comment;
  import Indent;
  import Section;

  # 
  # Returns the Fully Qualified name of the component
  #
  method get_fqn {} {
    set name  [$self get_name];
    set pkg   [$self find_ancestor "Package"];
    while { $pkg != "" } {
      if { [$pkg is_kind_of "Model"] } break;
      set mname [$pkg get_name];
      set name "$mname.$name";
      set pkg [$pkg find_ancestor "Package"];
    }    
    return $name;
  }

  # 
  # Returns the source directory for the component
  #
  method get_directory {} {  
    set pkg [$self find_ancestor "Package"];
    if { $pkg == "" } return;
    set src_dir [$pkg get_src_dir];
#    if { [sc_platform] == "macintosh" } {
#      append src_dir ":"; # XXX Bug in Mac TCL 7.5
#    }
    if { $src_dir == "" || ! [file exists $src_dir] } { 
      set src_dir [$pkg ask_src_dir];
      if { [sc_platform] == "macintosh" } {
        append src_dir ":"; # XXX Bug in Mac TCL 7.5
        if { [file exists [file join $src_dir "x.src"]] } {
          rm [file join $src_dir "x.src"]; # XXX Bug in PepperSeed / CD 1.1.3
                                           # XXX What about new ClassDesigner?
        }
      }
    }
    return $src_dir;
  }

  staticmethod find {namespace name kind} {
    set components [$namespace get_children "Component"];
    foreach component $components {
      if { [$component get_name] != $name } continue;
      if { [$component get_component_kind] != $kind } continue;
      return $component;
    }
    return "";
  }

  #
  # Prints the source code corresponding this component
  # to the appropriate file
  #
  method print {} {
    #
    # In java we generate code only for implementation 
    #
    if { [$self get_component_kind] != "implementation" } {
      return;
    }

    #
    # Look up package
    #
    set package [$self find_ancestor "Package"];
    if { $package == "" } return;
    if { ! [$package is_writable] } return;

    #
    # Look up component name and directory
    #
    set name [$self get_name];
    if { $name == "" } return;
    set dir [$self get_directory];
    if { $dir == "" } return;

    #
    # Generate file names
    #
    set impl_src [file join $dir "${name}.java"];
    set impl_bak [file join $dir "${name}.java.bak"];
    set impl_tmp [file join $dir "${name}.java.tmp"];

    #
    # Check if the implementation file can be written
    #
    if { [file exists $impl_src] && ![sc_isFileWritable $impl_src] } {
      Console warning "$name: $impl_src is write protected"; 
      return;
    }

    #
    # Find the namespace for the component (XXX I don't understand this?)
    #
    set namespace [$self find_ancestor "Namespace"];
    if { $namespace == "" } {
      Console warning "$name: component has no namespace";
      return;
    }
    
    # 
    # Find the implementation component corresponding to this source component
    #
    set impl_cmp [find $namespace $name "implementation"];
    if { $impl_cmp == "" } {
      Console warning "$name: component has no implemetation";
      return;
    }

    #
    # Verify the integrity of the component
    # 
    set num_errors [$self violations];
    if { $num_errors > 0 } {
      Console message "$num_errors error(s). Skipping file $impl_src";
      return;
    }
    
    #
    # Extract code sections from the existing implementation
    #
    Section extract $impl_src;

    #
    # Print the implementation
    #
    $impl_cmp print_implementation $impl_src $impl_bak $impl_tmp;

    # 
    # Forget_sections
    #
    Section init;

    #
    # Update the section ids
    #
    ^ $self set_csid "[$self get_name]_java"; # XXX
  }


  #
  # Verifies the integrity of derivations
  #
  method violations {{mode verbose}} {
    set derivations [$self get_mdlelm_derivations];
    set num_errors 0;
    foreach derivation $derivations {
      set mdl_elm [$derivation get_model_element];
      Console message "Verifying model element [$mdl_elm get_name]..."

      if { [$mdl_elm get_name] == "" } continue;

      switch [$derivation get_derivation_kind] {
	implementation { 
	  set num_errors [$mdl_elm violations $mode];
	}
      }

    }
    return $num_errors;
  }

  #
  # Generates implementation code
  #
  method print_implementation {src_name bak_name tmp_name} {
    Console message "Writing $src_name..."

    #
    # Generate the code to a temporary file
    #
    $self print_source $tmp_name;


    #
    # Copy Macintosh resource fork and create/type to the new file
    #
    if { [file exists $src_name] } {
      sc_copyResourceFork   $src_name $tmp_name;
      sc_copyCreatorAndType $src_name $tmp_name;
    } else {
      sc_setCreatorAndType  $tmp_name "CWIE" "TEXT";  # XXX
    }

    #
    # Create a backup
    #
    if { [file exists $src_name] } {
      if { [file exists $bak_name] } {
	sc_deleteFile $bak_name;
      }
      sc_renameFile $src_name $bak_name;
    }
    
    # 
    # Rename tmp file as source
    #
    sc_renameFile $tmp_name $src_name;
  }

  #
  # Prints out the source code to the given file
  #
  method print_source {filename} {
    
    Output open $filename;

    Indent init;

    # 
    # Component comment
    #
    Comment print begin;
    Comment print copyright;
    Comment print cvs;

    Comment print line "Component:  [$self get_name]";
    Comment print line "Package: [[$self find_ancestor Package] get_name]";
    Comment print line "File:    [$self get_name].java";
    Comment print line "";
    Comment print text "[$self get_description]";
    Comment print line "";

    Comment print end;

    Output newline;

    # 
    # Package directive
    #
    set package [$self find_ancestor "Package"];
    if { $package != {} && [$package get_kind] == "Package" } {
      Output print "package [$package get_fqn];";
    }

    #
    # Import clauses 
    #
    set dependencies [$self get_supplier_elms];
    foreach dependency $dependencies {
      if { [$dependency get_name] == {} } continue;
      Output print "import [$dependency get_fqn];";
    }
 
    #
    # Explicit declarations
    #

    Output print "[$self get_declarations]";

    Output newline;

    #
    # Classes and interfaces
    #
    set derivations [$self get_mdlelm_derivations];
    foreach derivation $derivations {
      set mdl_elm [$derivation get_model_element];

      if { [$mdl_elm get_name] == "" } continue; # No name, don't generate

      switch [$derivation get_derivation_kind] {
	implementation { $mdl_elm print; }
      }
    }

    #
    # Unused code
    #
    Section printUnused;

    #
    # Close output
    #
    Output close;
  }    
}

