#
# Interfac.tcl
#
# ClassDesigner Interface meta-object handling
#

package provide Interface 0.1;

class Interface {

  extends Element;

  import Output;
  import Indent;
  import Comment;
  import Console;
  import library;

  staticarray template;

  source [file join [library] TmplInfc.tcl];

  #
  # Returns the (recursive set of) interfaces this interface extends
  #
  method interfaces {} {
    set interfaces {};
    set generalizations [$self get_generalizations];
    foreach generalization $generalizations {
      set superclass [$generalization get_superclass];
      if { ![$superclass is interface] } continue;
      set interfaces \
	  [concat $interfaces [concat $superclass [$superclass interfaces]]];
    }
    return $interfaces;
  }

  #
  # Creates a virtual method which is mapped over elements of a list
  #
  staticmethod map {list code} {
    set result {};
    foreach self $list {
      lappend result [eval $code];
    }
    set result [join $result "\, "];
    return $result;
  }

  #
  # Verifies the integrity of the interface
  #
  # NOTE: Attributes or Operations may not be referenced before
  #       verifying the integrity of Interface!
  #       (Interface must be allowed to explicitly create
  #       Attributes and Operations and provide them
  #       with the proper Classifier)
  #
  method violations {{mode verbose}} {

    set num_errors 0;

    #
    # No superclasses allowed
    #
    set generalizations [$self get_generalizations];
    foreach generalization $generalizations {
      set superclass [$generalization get_superclass]; 
      if { $superclass == "" } continue;
      if { [$superclass is interface] } continue;
      $self error "interface extending a class ( [$superclass get_name] )" \
	  $mode;
      incr num_errors;
    }

    #
    # Non-abstract methods (and non-static/non-final fields)
    # are not considered an integrity violation;
    # they are silently forced to "abstract" (to final and static).
    # See Operation and Attribute.
    #
    set attributes [$self get_attributes];
    foreach attribute $attributes {
      new Attribute $attribute $self;
    }
    set operations [$self get_operations];
    foreach operation $operations {
      new Operation $operation $self;
    }

    return $num_errors;
  }

  #
  # Prints the interface
  #
  method print {{innerclass 0}} {
    set name [$self get_name];
    
    #
    # Print description
    #
    if { ! $innerclass } {
      Comment print description "$name\n\n[$self get_description]";
    }

    #
    # Print beginning
    #
    set expander [new Expander];

    if { $innerclass  } {
      $expander bind visibility "";
    } else {
      $expander bind visibility "[$self prop visibility]";
    }

    $expander bind abstract   "[$self prop abstract]";
    $expander bind final      "[$self prop final]";
    $expander bind name	      "$name";
    $expander bind interfaces "[map [$self interfaces] {$self get_name}]";

    Output print [$expander expand $template(interface)];
 
    Indent more;

    #
    # Print attributes
    # 
    set attributes [$self get_attributes];
    foreach attr $attributes {
      $attr print [$self get_name];
    }

    #
    # Print operations
    #
    set operations [$self get_operations];
    foreach operation $operations {
      $operation print $self; # self passed to Operation's init method
    }

    #
    # Print end
    #
    Indent less;

    Output print [$expander expand $template(end)];

    delete $expander;
  }
}

