#
# TclClass.tcl
#
# OO programming with TCL
#
# Each class is an interpreter
#  - static fields are standard interp global vars
#  - static methods are standard procs, with appropriate "global" clause
#  - instances are named .N, where N is an number
#  - instance fields are stored in .N(name)
#  - instance method are procs with extra self argument and
#    appropriate upvars for accessing .N(name) fields
#  - sc_elem is an special instance field
#  - each instance is also a proc
# 

package provide TclClass 0.1;


#
# Creates a new class of the given name (and baseclass)
#

proc class {classname code} {

  set class [interp create];

  $class eval "set .class $classname";	# Classname for the interp

  $class eval {
    set       .count  0;		# Object count
    array set .      {};		# Instance fields names & initial vals
  }

  #
  # Access to upper level
  #
  $class alias ^ eval;				

  #
  # Creates a new symbol 
  #
  $class eval "
    proc .gensym {} {
      global .count;

      return \"${classname}_\[incr .count\]\";
    }
  "

  # 
  # Tries to alias an symbol from a different class' namespace
  #
  $class eval "
    proc .linksym {otherclass symbol} {
      ^ interp alias $class \$symbol \[^ \$otherclass .linksource \$symbol\];
    }
  "

  $class eval "
    proc .linksource {symbol} {
      return \"$class \$symbol\";
    }
  "
 
  #
  # Keywords that work within a class
  #
  $class eval {
    #
    # Create and delete objects 
    #
    proc new {class args} {
      ^ $class .new $args;
    }

    proc delete {object args} {
      ^ [.classname $object] .delete $object $args;
    }

    proc .new {{name {}} args} {
      global .;

      if { $name == {} } {
	set name [.gensym];
      }

      proc $name {method args} "eval \$method $name \$args";
      global ".$name";
      foreach var [array names .] {
	set ".[set name]($var)" [set .($var)];
      }
      if { [info proc init] == "init" } {
	eval init $name $args;
      }
      return $name;
    }

    proc .delete {name args} {
      global $name;
      if { [info proc finalize] == "finalize" } {
        eval finalize $name $args;
      }
      array set ".$name" {};
      rename $name {};
    }

    #
    # Aliases unknown symbols from other classes
    #
    if { [info command unknown] == "unknown" } {
      rename unknown .unknown;
    } else {
      proc .unknown {name args} {
	return -code error "invalid command name \"$name\"";
      }
    }

    proc unknown {object args} {
      set class [.classname $object];
      if { $class != {} } {
	if { [^ $class info proc $object] == $object } {
	  .linksym $class $object;
	  return [eval $object $args];
	}
      }
      eval .unknown $object $args;
    }

    proc .classname {object} {
      set lst [split $object _];
      if { [llength $lst] == 2 } {
	return [lindex $lst 0];
      }
      return {};
    }

    #
    # Defines a new fields (with given initial value)
    #
    proc field {name {value {}}} {
      global . ;
      set .($name) $value;
    }

    #
    # Defines a new static field
    #
    proc staticfield {name {value {}}} {
      global $name;
      set $name $value;
    }

    # 
    # Defines a new static array
    #
    proc staticarray {name {valuelist {}}} {
      global $name;
      array set $name $valuelist;
      set ${name}(___) x;
      unset ${name}(___);
    }

    #
    # Returns an expression that binds field names, used by method
    #
    proc .fieldlist {} {
      global .

      set fieldlist {};
      foreach name [array names .] {
        append fieldlist "upvar #0 \".\[set self\]($name)\" $name; ";
      }
      append fieldlist [.staticlist];
      return $fieldlist;
    }

    #
    # Returns an expression that binds static field names
    #
    proc .staticlist {} {
      global .

      set fieldlist {};
      foreach name [info globals {[A-Za-z_]*}] {
        if { [string match "tcl_*" $name] } continue;
        append fieldlist "global $name; ";
      }
      return $fieldlist;
    }

    #
    # Creates a new method with given name, arguments and body
    # If a method overrides another, the overridden method can still be
    # accessed using notation <superclass name>.<method>
    #
    proc method {name argv body} {
      global .class;
      proc "${.class}.$name" [concat "self" $argv] \
	  [concat [.fieldlist] $body];
      proc $name [concat "self" $argv] [concat [.fieldlist] $body];
    }

    # 
    # Creates a static method
    #
    proc staticmethod {name argv body} {
      proc $name $argv                [concat [.staticlist] $body];
    }

    #
    # Expands an argument list with defaults
    #
    proc .defexp {proc args} {
      set ret {};
      foreach arg $args {
	if { [info default $proc $arg def] } {
	  lappend ret [list $arg $def];
	} else {
	  lappend ret $arg;
	}
      }
      return $ret;
    }

    #
    # Inherits fields and methods from some other class
    # (smalltalk semantics with static fields, i.e. separate)
    #
    proc extends {class} {
      global .

      set procs [^ $class info procs {{[A-Za-z_]*}}];
      foreach proc $procs {
	switch $proc {
	  new		-
	  delete	-
	  field		-
	  staticfield	-
	  staticarray	-
	  method	-
	  staticmethod	-
	  extends	-
	  import	continue;
	  default {
	    proc $proc "[^ $class .defexp $proc [^ $class info args $proc]]" \
		"[^ $class info body $proc]";
	  }
	}
      }

      set globals [^ $class info globals {{[A-Za-z_]*}}];
      foreach global $globals {
	global $global
        if { $global == "env" } continue;
        if { [string match "tcl_*" $global] } continue;
        if { [array exists $global] } continue;
	if { [^ $class array exists $global] } {
	  array set $global [^ $class array get_$global];
	} else {
	  set $global [^ $class "expr {\$$global};"];
	}
      }

      set fields  [^ $class array names .];
      foreach field $fields {
        set .($field) [ ^ $class "set .($field)"];
      }
    }
  }

  # 
  # Allows class level references to another class
  #
  $class eval "
    proc import {class} { ^ $class alias \$class \$class; }
  ";

  #
  # Create the class command
  # 
  proc $classname {method args} "$class eval \$method \$args"

  #
  # Evaluate class body
  #

  $class eval $code;
}

