/**
    Copyright (C) 1997 Pekka Nikander & Juha Prssinen.
    All rights Reserved.

    XXX   %Z%%M% version %I% date:%G%

    Module:  Controller (implementation)
    Package: conduits
    File:    Controller.java


*/
package fi.hut.tcm.jacob.conduits;
/**
    Controller
    
    Controller is the abstract base class for Visitors that perform
    surgical operations for the conduit graph. That is, a Controller will
    travers the conduit graph, looking for a proper place to e.g. install a
    new conduit or remove an existing one. Once the right location is
    found, the Controller performs the desired operation. 
*/
public abstract class Controller
  implements Visitor {
    /**
        The conduit this Controller is currently at, or the conduit this
        Controller will next visit.
    */
    protected Conduit curr;
    /**
        The conduit this Controller last visited. Installed to
        <code>origin</code> by <code>init</code>.
    */
    protected Conduit prev;
    protected Conduit next;

    /**
        Creates a new empty Controller. Provided for subclass constructors.
    */
    protected Controller() {
    //+ Controller::Controller( )
    //- Controller::Controller( )
    }

    public void at(Mux mux) {
    //+ Controller::at(Mux)
      if (destination(mux)) {
	operate();
      }
    //- Controller::at(Mux)
    }

    public void at(Session session) {
    //+ Controller::at(Session)
      if (destination(session)) {
	operate();
      }
    //- Controller::at(Session)
    }

    public void at(Adaptor adaptor) {
    //+ Controller::at(Adaptor)
      if (destination(adaptor)) {
	operate();
      }
    //- Controller::at(Adaptor)
    }

    public void at(ConduitFactory factory) {
    //+ Controller::at(ConduitFactory)
      if (destination(factory)) {
	operate();
      }
    //- Controller::at(ConduitFactory)
    }

    public Conduit getPrev() {
    //+ Controller::getPrev( )
      return prev;
    //- Controller::getPrev( )
    }

    public Conduit getNext() {
    //+ Controller::getNext( )
      return next;
    //- Controller::getNext( )
    }

    public void setNext(Conduit conduit) {
    //+ Controller::setNext(Conduit)
      next = conduit;
    //- Controller::setNext(Conduit)
    }

    /**
        Initializes the start and origin attributes. Typically called by a
        subclass method to initialize a controller for traverse.
    */
    protected void init(Conduit start, Conduit origin) {
    //+ Controller::init(Conduit, Conduit)
      curr = start;
      prev = origin;
    //- Controller::init(Conduit, Conduit)
    }

    protected void reset() {
    //+ Controller::reset( )
      this.init(null, null);
    //- Controller::reset( )
    }

    /**
        Determines if the candidate conduit is the right locationto perform
        the operation. May be overridden by a subclass. The default
        implementation returns true on a Mux, false otherwise.
    */
    protected boolean destination(Conduit candidate) {
    //+ Controller::destination(Conduit)
      return (candidate instanceof Mux);
    //- Controller::destination(Conduit)
    }

    /**
        Called when the controller reaches the destination.
        Must be overridden by a subclass.
    */
    protected abstract void operate();

    /**
        Traverses the conduit graph, looking for destination. Starts at
        (origin,start). The destination is reached when the destination()
        method returns true within the appropriate at() method.
    */
    public void run() {
    //+ Controller::run( )
      while (curr != null) try {
	next = null;
	curr.accept(this); // May modify next through setNext
	if (destination(curr)) break; // XXX
	prev = curr;
	curr = next;
      } catch (ConductingException ex) {
	ex.printStackTrace();	// XXX
	throw new InternalError(ex.getMessage());
      }
    //- Controller::run( )
    }

}; //Controller

