/**
  XXX   %Z%%M% version %I% date:%G%
  Component:  BaseConduit
  Package: conduits
  File:    BaseConduit.java
*/

package fi.hut.tcm.jacob.conduits;

/**
  BaseConduit
  
  A BaseConduit is an invisible abstract class that implements the common 
  behaviour for the real conduits. It is not part of the externally visible
   model but an internal implementation class.
*/
public abstract class BaseConduit
  extends Object
  implements Conduit, Cloneable {

  /**
  */
  protected static final int SIDE_NUMBER = 2;
  
  /**
  */
  protected Side[] sides;
  
  /**
  */
  protected BaseConduit() {
  //+ BaseConduit::BaseConduit( )
    initSides();
  //- BaseConduit::BaseConduit( )
  }
  
  /**
  */
  protected abstract void accept(Visitor visitor, Side from)
    throws ConductingException, IllegalArgumentException ;
  
  /**
  */
  protected void initSides() {
  //+ BaseConduit::initSides( )
    BaseSide a, b;
    
    a = new BaseSide();
    b = new BaseSide();
    a.opposite = b;
    b.opposite = a;
    
    sides = new BaseSide[initialSideCount()];
    sides[A] = a;
    sides[B] = b;
  //- BaseConduit::initSides( )
  }
  
  /**
  */
  protected int initialSideCount() {
  //+ BaseConduit::initialSideCount( )
    return SIDE_NUMBER;
  //- BaseConduit::initialSideCount( )
  }
  
  /**
  */
  public Side getSide(int index) {
  //+ BaseConduit::getSide(int)
    if (index >= getSideCount()) 
      throw new ArrayIndexOutOfBoundsException();
    return sides[index];
  //- BaseConduit::getSide(int)
  }
  
  /**
  */
  public int getSideCount() {
  //+ BaseConduit::getSideCount( )
    return sides.length;
  //- BaseConduit::getSideCount( )
  }
  
  /**
  */
  public Object clone()
    throws CloneNotSupportedException  {
  //+ BaseConduit::clone( )
    BaseConduit clone = (BaseConduit)super.clone();
    
    clone.initSides();
    return clone;
  //- BaseConduit::clone( )
  }
  
  /**
    BaseSide
    
    
  */
  class BaseSide
    extends Object
    implements Side {
  
    /**
    */
    transient Side attached;
    
    /**
    */
    BaseSide opposite;
    
    /**
      If this side belongs to a session, then it
      will set the messenger for the message
      passing through. For non-sessions, the messenger
      is null.
    */
    private Messenger msgr;
    
    /**
    */
    public void attach(Side attached)
      throws SideException, IllegalArgumentException  {
    //+ BaseSide::attach(Side)
      if (attached == null)
        throw new IllegalArgumentException("attach to null");
      if (attached == opposite)
        throw new IllegalArgumentException("attach to opposite");
      if (attached == this)
        throw new IllegalArgumentException("attach to itself");

      if (this.attached == attached) 
        return; // Allow idempontent calls

      if (this.attached != null)
        throw new OccupiedSideException("attach on attached");

      this.attached = attached;

      // XXX pnr: I'm not quite sure if we should do this?
      attached.attach(this);      
    //- BaseSide::attach(Side)
    }
    
    /**
       Detaches this side from the side this side is connected to.
       
       Will not complain if this side is connected to a side which
       is not connected to this side. Also will not complain if
       this side is not connected to anything. Will only cause
       an exception if this side is connected to a side which
       is connected to a side other than this one.
    */
    public Side detach() {
    //+ BaseSide::detach( )
      if (this.attached == null)
        return null; // Allow idempontent calls

      Side old_attached = attached;
      attached = null;

      // XXX pnr: I'm not quite sure if we should do this?
      Side s = old_attached.detach();
      if ((s != null) && (s != this))
        throw new IllegalStateException("crossreferences on detach"); 
      
      return old_attached;
    //- BaseSide::detach( )
    }
    
    /**
    */
    public void accept(Visitor visitor)
      throws ConductingException  {
    //+ BaseSide::accept(Visitor)
	// XXX DEBUG:
      //System.out.print("Visitor at " + this + " belonging to ");
      //System.out.println( BaseConduit.this + " [BaseSide.accept()]");
	BaseConduit.this.accept(visitor, this);
    //- BaseSide::accept(Visitor)
    }
    
    /**
    */
    public Conduit getConduit() {
    //+ BaseSide::getConduit( )
      return BaseConduit.this;
    //- BaseSide::getConduit( )
    }
    
    /**
    */
    public Side getNext() {
    //+ BaseSide::getNext( )
      return opposite != null? opposite.attached: null;
    //- BaseSide::getNext( )
    }
    
    /**
    */
    public Side getAttached() {
    //+ BaseSide::getAttached( )
      return attached;
    //- BaseSide::getAttached( )
    }
    
    /**
    */
    protected Side getOpposite() {
    //+ BaseSide::getOpposite( )
      return opposite;
    //- BaseSide::getOpposite( )
    }
    
    /**
    */
    public Object clone()
      throws CloneNotSupportedException  {
    //+ BaseSide::clone( )
      throw new CloneNotSupportedException("framework sides can't be directly cloned");
    //- BaseSide::clone( )
    }
    
    /**
    */
    BaseSide() {
    //+ BaseSide::BaseSide( )
      msgr = null;
    //- BaseSide::BaseSide( )
    }
    
    /**
    */
    BaseSide(Messenger msgr) {
    //+ BaseSide::BaseSide(Messenger)
      this.msgr = msgr;
    //- BaseSide::BaseSide(Messenger)
    }
    
    /**
    */
    public void setMessenger(Messenger msgr) {
    //+ BaseSide::setMessenger(Messenger)
      this.msgr = msgr;
    //- BaseSide::setMessenger(Messenger)
    }
    
  } // BaseSide
  
} // BaseConduit

// +*************** UNUSED CODE ***************

// -*************** UNUSED CODE ***************
