/**
  Copyright (c) 1998 Pekka Nikander & Juha Parssinen
  All rights reserved.
  XXX   %Z%%M% version %I% date:%G%
  Module:  ConduitFactory
  Package: conduits
  File:    ConduitFactory.java
*/
package fi.hut.tcm.jacob.conduits;
/**
  ConduitFactory
  
  A ConduitFactory creates and installes new Sessions or other Conduit 
  structures. Whenever a MessageTransporter arrives at a ConduitFactory, it
  calls the facture() method. This makes the factory to create a new 
  conduit structure, and install it at appropriate locations within the 
  existing conduit graph.
*/
public final class ConduitFactory
  extends BaseConduit
  implements FactoryEventListener {

  /**
  */
  private Conduit conduitPrototype;
  
  /**
  */
  protected static final int SIDE_NUMBER = 2;
  
  /**
  */
  protected static final int ENDPOINT_NUMBER = 2;
  
  /**
  */
  protected static final int FIRST_ENDPOINT_INDEX = SIDE_NUMBER;
  
  /**
  */
  protected static final int TOTAL_SIDE_NUMBER = FIRST_ENDPOINT_INDEX + ENDPOINT_NUMBER;
  
  /**
    Creates a new ConduitFactory.
    @param conduitPrototype    A Conduit to be cloned if not specified
           in the message given to facture().
    @param endpoints           Sides to which to attach the cloned
           Conduit if not specified in the message given to facture().
           If a MuxSide is given, the side to use is asked from
	   the Mux, passing the message to the Mux so that it
           can determine which side to use, or create one if a suitable
           side doesn't already exist. The length of the array must
	   be ENDPOINT_NUMBER (2).
  */
  public ConduitFactory(Conduit conduitPrototype, Side[] endpoints) {
  
    //+ ConduitFactory::ConduitFactory(Conduit, Side[])
    super();
    
    this.conduitPrototype = conduitPrototype;

    if (sides.length != FIRST_ENDPOINT_INDEX + endpoints.length)
      throw new InternalError("not implemented yet");
    for (int i = 0; i < endpoints.length; i++)
      sides[FIRST_ENDPOINT_INDEX + i] = endpoints[i];
    //- ConduitFactory::ConduitFactory(Conduit, Side[])
  }
  
  
  /**
  */
  public Side getEndpoint(int index) {
  
    //+ ConduitFactory::getEndpoint(int)
    return sides[FIRST_ENDPOINT_INDEX + index];
    //- ConduitFactory::getEndpoint(int)
  }
  
  
  /**
  */
  public void setEndpoint(int index, Side endpoint) {
  
    //+ ConduitFactory::setEndpoint(int, Side)
    sides[FIRST_ENDPOINT_INDEX + index] = endpoint;
    //- ConduitFactory::setEndpoint(int, Side)
  }
  
  
  /**
  */
  public Conduit getPrototype() {
  
    //+ ConduitFactory::getPrototype( )
    return conduitPrototype;
    //- ConduitFactory::getPrototype( )
  }
  
  
  /**
  */
  public void setPrototype(Conduit prototype) {
  
    //+ ConduitFactory::setPrototype(Conduit)
    conduitPrototype = prototype;
    //- ConduitFactory::setPrototype(Conduit)
  }
  
  
  /**
  */
  protected final void accept(Visitor visitor, Side from)
    throws ConductingException, IllegalArgumentException  {
  
    //+ ConduitFactory::accept(Visitor, Side)
    visitor.at(this, from);
    //- ConduitFactory::accept(Visitor, Side)
  }
  
  
  /**
  */
  protected int initialSideCount() {
  
    //+ ConduitFactory::initialSideCount( )
    return TOTAL_SIDE_NUMBER;
    //- ConduitFactory::initialSideCount( )
  }
  
  
  /**
  */
  public int getSideCount() {
  
    //+ ConduitFactory::getSideCount( )
    return SIDE_NUMBER;
    //- ConduitFactory::getSideCount( )
  }
  
  
  /**
    Creates a new conduit by cloning the prototype. The newly created 
    conduit is wrapped into two installers, created from the Installer 
    prototypes, which are then sent to sides A and B. The message is 
    included into the Installer; it can be used to determine where to 
    install the newly created conduit.
  */
  public void facture(Message message) {
  
    //+ ConduitFactory::facture(Message)
    Conduit prototype = null;
    Side[] endpoints;
    
    Messenger messenger = message.getMessenger();
    if (messenger instanceof InstallMessenger) {
      // XXX DEBUG
      System.out.println("ConduitFactory.facture(message): InstallMessenger detected!");
      InstallMessenger im = (InstallMessenger)messenger;
      
      prototype = im.getPrototype(message);
      endpoints = new Side[im.getEndpointCount(message)];
      for (int i = 0; i < endpoints.length; i++)
        endpoints[i] = im.getEndpoint(message, i);
    } else {
      endpoints = new Side[ENDPOINT_NUMBER];
    }
    
    if (prototype == null)
      prototype = conduitPrototype;  // This can also be null
    
    for (int i = 0; i < ENDPOINT_NUMBER && i < endpoints.length; i++) {
      if (endpoints[i] == null) {
        endpoints[i] = sides[FIRST_ENDPOINT_INDEX + i];
      }
    }
    
    facture(message, prototype, endpoints);
    //- ConduitFactory::facture(Message)
  }
  
  
  /**
  */
  private void facture(Message message, Conduit prototype, Side[] endpoints) {
  
    //+ ConduitFactory::facture(Message, Conduit, Side[])
    Conduit clone = null;
    int endpointCount;
    MuxSide mx;
    boolean success = false;
    
    try {
      clone = (Conduit)prototype.clone();
      endpointCount = clone.getSideCount();
      
      if (endpointCount != endpoints.length)
        throw new NoSuchSideException("endpoint count mismatch");
        
      for (int i = 0; i < endpointCount; i++) {
        if (endpoints[i] instanceof MuxSide) {
          Mux mux = (Mux)(endpoints[i].getConduit());
          Side side = mux.getSide(message);
          if (side == null)
            side = mux.createSide(message);
          endpoints[i] = side;
        } 
        endpoints[i].attach(clone.getSide(i));
      }

      success = true;
      
    } catch (Exception e) {
      System.out.println("XXX ConduitFactory.facture failed:" + e.getMessage());
      e.printStackTrace();

      //System.exit(1);
    }


    // Send notification to the client.
    //
    
    // We are free to include anything we want in the message,
    // and in this case we'll include a reference to the cloned object.
    // If the client then wants to change connections to the object
    // by itself, it can do so.
    Object nContents = (Object)clone;
    
    // The request specified in 'message' was successful.
    NotifyOob nOob = new NotifyOob(message,
				   success?NotifyMessenger.SUCCESS:
				   NotifyMessenger.FAILURE);
    
    Message nMessage = Message.newMessage(nContents,
					  Notifier.instance,
					  nOob);

    // Have the message trasported back to the direction from where
    // the original message came from.
    MessageTransporter nMt =
      MessageTransporter.newMessageTransporter(message.getSourceSide().
					       getAttached(), nMessage);
    SimpleScheduler.instance.makeRunnable(nMt);
    
    //- ConduitFactory::facture(Message, Conduit, Side[])
  }
  
  
  /**
  */
  public Object clone()
    throws CloneNotSupportedException  {
  
    //+ ConduitFactory::clone( )
    return super.clone();
    // The conduitPrototype must not have any external references,
    // therefore, it need not be cloned.
    
    // The sides to attach are stored in the sides[] array.
    // Therefore the code in Protocol.clone will take care of updating them.
    //- ConduitFactory::clone( )
  }
  
  
} // ConduitFactory


