/**
  Copyright (c) 1998 Pekka Nikander & Juha Parssinen
  All rights reserved.
  XXX   %Z%%M% version %I% date:%G%
  Component:  Mux
  Package: conduits
  File:    Mux.java
*/
package fi.hut.tcm.jacob.conduits;
import java.util.Hashtable;
public final class Mux
  extends BaseConduit {

  /**
  */
  private Accessor accessor;
  
  /**
  */
  private int sideCount;
  
  /**
  */
  public Mux(Accessor accessor) {
  //+ Mux::Mux(Accessor)
    this.accessor = accessor;
  //- Mux::Mux(Accessor)
  }
  
  /**
  */
  protected void accept(Visitor visitor, Side from)
    throws ConductingException, IllegalArgumentException  {
  //+ Mux::accept(Visitor, Side)
    visitor.at(this, from);
  //- Mux::accept(Visitor, Side)
  }
  
  /**
    Attaches to the message the key corresponding
    to the conduit where the message arrived from, or
    decodes the key from the message and routes it appropriately.
  */
  public void mux(Message message) {
  //+ Mux::mux(Message)
    MuxSide side = (MuxSide)message.getSource();
    
    // Perform encoding of key, or removal of key if needed
    side.code(message);
    
  //- Mux::mux(Message)
  }
  
  /**
  */
  protected void initSides() {
  //+ Mux::initSides( )
    BaseSide a, b0;
   
    // Both A and B0 try to demux messages.  However, their behaviour
    // differs if demuxing fails.  A sends to B0 (the opposite),
    // B0 sends to A (the opposite).  @see DemuxSide.getNext(Message) 
    a  = new DemuxingSide();
    b0 = new DemuxingSide();
    
    // Connect a to b0 and vice versa.
    a.opposite  = b0;
    b0.opposite = a;
    
    sides = new MuxSide[initialSideCount()];
    sides[A] = a;
    sides[B] = b0;
    
    sideCount = sides.length;
  //- Mux::initSides( )
  }
  
  /**
  */
  public int getSideCount() {
  //+ Mux::getSideCount( )
    return sideCount;
  //- Mux::getSideCount( )
  }
  
  /**
  */
  protected Side getSide(Message message) {
  //+ Mux::getSide(Message)
    
    if (message == null)
      throw new IllegalArgumentException("null message");
      
    Object key = accessor.getKey(message);

    for (int i = initialSideCount(); i < sideCount; i++)
      if (((MuxingSide)sides[i]).key.equals(key)) 
        return sides[i];
    return null;
  //- Mux::getSide(Message)
  }
  
  /**
  */
  protected Side createSide(Message message) {
  //+ Mux::createSide(Message)
    if (message == null)
      throw new IllegalArgumentException("null message");
    Object key = accessor.getKey(message);
    return createSide(key);
  //- Mux::createSide(Message)
  }
  
  /**
  */
  public Side createSide(Object key) {
  //+ Mux::createSide(Object)

    for (int i = initialSideCount(); i < sideCount; i++)
      if (((MuxingSide)sides[i]).key.equals(key))
        throw new IllegalArgumentException("side already exists");
      
    // Create new side and store it 
    MuxSide side = new MuxingSide(key, (BaseSide)sides[A]);
    if (sideCount == sides.length) {
      // Grow array
      BaseSide[] n = new BaseSide[sides.length * 2];
      System.arraycopy(sides, 0, n, 0, sides.length);
      sides = n;
    }
    
    sides[sideCount++] = side;
    return side;

  //- Mux::createSide(Object)
  }
  
  /**
  */
  protected void deleteSide(Side side) {
  //+ Mux::deleteSide(Side)
    if (side == null)
      throw new IllegalArgumentException("null side");
    if (!(side instanceof MuxingSide))
      throw new IllegalArgumentException("not MuxingSide");
      
    MuxingSide muxside = (MuxingSide)side;
    for (int i = initialSideCount(); i < sideCount; i++) {
      if (sides[i] == muxside) {
        System.arraycopy(sides, i+1, sides, i, sideCount - i-1);
        sideCount--;
        return;
      }
    }
    
    throw new IllegalArgumentException("side not found");
  //- Mux::deleteSide(Side)
  }
  
  /**
  */
  public Object clone()
    throws CloneNotSupportedException  {
  //+ Mux::clone( )
    Mux clone = (Mux)super.clone();
        
    for (int i = initialSideCount(); i < sideCount; i++) {
      clone.sides[i] = new MuxingSide(((MuxingSide)sides[i]).key, (BaseSide)(clone.sides[A]));
    }
    return clone;
  //- Mux::clone( )
  }

  /**
  */
  class MuxingSide extends BaseSide implements MuxSide {
    Object key;
    
    protected MuxingSide(Object key, BaseSide opposite) {
      //+ MuxingSide::MuxingSide(Object, Side)
      this.key = key;
      this.opposite = opposite;
      //- MuxingSide::MuxingSide(Object, Side)
    }
    
    public void code(Message message) {
      //+ MuxingSide::code(Message)
      accessor.setKey(message, key);
      //- MuxingSide::code(Message)
    }
    public Side getNext(Message message) {
      //+ MuxingSide::getNext(Message)
      return getNext();
      //- MuxingSide::getNext(Message)
    }
  }

  /**
  */
  class DemuxingSide extends BaseSide implements MuxSide {
    public void code(Message message) {
      //+ DemuxingSide::code(Message)
      //System.out.println("code(): unsetting key");
      accessor.unsetKey(message);
      //+ DemuxingSide::code(Message)
    }
    
    public Side getNext(Message message) {
      //+ DemuxingSide::getNext(Message)
      Object key = accessor.getKey(message);
      BaseSide opposite = null;
      for (int i = initialSideCount(); i < sideCount; i++) 
	if (((MuxingSide)sides[i]).key.equals(key)) 
	  //if (((MuxingSide)sides[i]).key == key)
	  {
	    opposite = (BaseSide)sides[i];
	    //System.out.println("DemuxingSide: found match " + i);
	    break;
	  }
      if (opposite == null) {
	//System.out.println("opposite == null");
        return getNext(); // Fallback to default
      }
      return opposite.attached;
      //- DemuxingSide::getNext(Message)
    }
  }
} // Mux

// +*************** UNUSED CODE ***************
// -*************** UNUSED CODE ***************
