/**
  Copyright (c) 1998 Pekka Nikander & Juha Parssinen
  All rights reserved.
  XXX   %Z%%M% version %I% date:%G%
  Module:  Protocol
  Package: conduits
  File:    Protocol.java
*/
package fi.hut.tcm.jacob.conduits;
/**
  Protocol
  
  A Protocol represents a larger, identifyable conduit structure. It allows
   several conduits to be used as a single unit. For example, a 
  ConduitFactory can create a whole substructure instead of a single 
  Session. To the traversing Visitors the Protocol is (almost) invisible. 
  That is, a Protocol is only a structural element, not a behavioural 
  element.
*/
public abstract class Protocol
  extends BaseConduit {

  /**
  */
  public static final int INSIDE_A = 2;
  
  /**
  */
  public static final int INSIDE_B = 3;
  
  /**
  */
  private static final int SIDE_NUMBER = 4;
  
  /**
  */
  protected Protocol() {
  
    //+ Protocol::Protocol( )
    //- Protocol::Protocol( )
  }
  
  
  /**
  */
  protected final void accept(Visitor visitor, Side from)
    throws ConductingException, IllegalArgumentException  {
  
    //+ Protocol::accept(Visitor, Side)
    // Cut through; alternative would be NOP.
    from.getNext().accept(visitor);
    //- Protocol::accept(Visitor, Side)
  }
  
  
  /**
  */
  protected void initSides() {
  
    //+ Protocol::initSides( )
    super.initSides();
    
    sides[INSIDE_A] = new BaseSide();
    sides[INSIDE_B] = new BaseSide();
    
    ((BaseSide)sides[A]).opposite = (BaseSide)sides[INSIDE_A];
    ((BaseSide)sides[B]).opposite = (BaseSide)sides[INSIDE_B];
    ((BaseSide)sides[INSIDE_A]).opposite = (BaseSide)sides[A];
    ((BaseSide)sides[INSIDE_B]).opposite = (BaseSide)sides[B];
    //- Protocol::initSides( )
  }
  
  
  /**
  */
  protected int initialSideCount() {
  
    //+ Protocol::initialSideCount( )
    return SIDE_NUMBER;
    //- Protocol::initialSideCount( )
  }
  
  
  /**
  */
  public Object clone()
    throws CloneNotSupportedException  {
  
    //+ Protocol::clone( )
    // This will create a new object, and two new sides for it, A and B.
    // Those sides will not be connected to any other Protocol.
    // It would seem that the inside sides are not created, however.
    Protocol clone = (Protocol)super.clone();
    
    /*
     * Perform a graph copy for the inside of the Protocol
     */
    java.util.Hashtable copied = new java.util.Hashtable();

    for (int i = 0; i < sides.length; i++)
      // Initialize a list of those sides which need not be copied
      // by graphcopy().
      // clone.sides[INSIDE_A] and clone.sides[INSIDE_B] will be
      // assigned values below.
      // Protocols do not have any more sides than the four
      // already mentioned.
      // (Muxen (only?) can have more than four sides.)
      copied.put(sides[i], clone.sides[i]);
      
    // Recursively copy reachable contents
    ((BaseSide)(clone.sides[INSIDE_A])).attached = graphcopy(copied, ((BaseSide)sides[INSIDE_A]).attached);
    ((BaseSide)(clone.sides[INSIDE_B])).attached = graphcopy(copied, ((BaseSide)sides[INSIDE_B]).attached);

    return clone;
    //- Protocol::clone( )
  }
  
  
  /**
  */
  private Side graphcopy(java.util.Hashtable copied, Side side)
    throws CloneNotSupportedException  {
  
    //+ Protocol::graphcopy(java.util.Hashtable, Side)
    // If this side was not set, terminate recursion
    // (Side not connected to one of some other component's sides.)
    if (side == null)
      return null;
    
    // If this side is already copied, terminate recursion    
    if (copied.containsKey(side))
      return (Side)copied.get(side);
      
    // If this side is not a BaseSide, try to clone
    if (!(side instanceof BaseSide)) {
      return (Side)side.clone(); // Most probably will fail
    }

    /*
     * Ok, we have now a new side that has not been copied yet.
     * Create a copy of the enclosing conduit, and insert
     * all sides of it to the hashtable.  It is important
     * to first insert them in the hashtable in order to prevent
     * multiple copies of the same conduits.  Thereafter loop
     * over the new sides, and recursively copy their next sides.
     */
     
    // Create conduit copy (this will create copies of all sides, unconnected)
    BaseConduit oc = (BaseConduit)(((BaseSide)side).getConduit());
    // (Seems that will only create copies of the _outer_ sides,
    // not the inner ones.)
    BaseConduit nc = (BaseConduit)oc.clone(); // does the copying
    
    // Create correspondence to terminate recursion
    for (int i = 0; i < oc.sides.length; i++) {
      copied.put(oc.sides[i], nc.sides[i]);
    }
    
    // Created sides have not been connected to anything yet.
    // Second round: recursive copying / connecting of sides
    for (int i = 0; i < oc.sides.length; i++) {
      // Possible problem here: nc.sides[2] and nc.sides[3] are null.
      ((BaseSide)(nc.sides[i])).attached = graphcopy(copied, ((BaseSide)(oc.sides[i])).attached);
    }
        
    // Return copy of the original
    return (Side)copied.get(side);
    
    //- Protocol::graphcopy(java.util.Hashtable, Side)
  }
  
  
} // Protocol
