/**
  Copyright (c) 1998 Pekka Nikander & Juha Parssinen
  All rights reserved.
  XXX   %Z%%M% version %I% date:%G%
  Module:  TimerThread
  Package: conduits
  File:    TimerThread.java
*/
package fi.hut.tcm.jacob.conduits;
import java.lang.Thread;
import fi.hut.tcm.jacob.utility.Queue;
import fi.hut.tcm.jacob.utility.OrderedQueue;
/**
  TimerThread
  
  TimerThread wakes up at appropriate times to check if there are anybody 
  to wake up. When TimerThread is created it creates queue for timeouts.
*/
public final class TimerThread
  extends Thread {

  /**
  */
  public static TimerThread instance = new TimerThread( );
  
  /**
  */
  private OrderedQueue timerQueue;
  
  /**
  */
  private Scheduler scheduler;
  
  /**
    The constructor of the TimerThread
  */
  private TimerThread() {
  
    //+ TimerThread::TimerThread( )
    timerQueue = new OrderedQueue(TimeoutQuery.instance);
    scheduler = SimpleScheduler.instance;
    this.start();
    //- TimerThread::TimerThread( )
  }
  
  
  /**
    Runs in a loop, checking for expired timeouts. All expired timeouts are
     moved to RunQueue. If there are any Timeouts left, calculates the next
     wakeup time and enters <code>wait</code>.
  */
  public synchronized void run() {
    Timeout t;
    long now;
    
    //+ TimerThread::run( )
    while (true) {
      t = (Timeout)timerQueue.take(); // blocks if empty
      now = System.currentTimeMillis();
      if (t.expirytime > now) try {
         wait(t.expirytime - now); // Wait until expired or woken
      } catch (InterruptedException e) {}
      if (t.expirytime == 0) {
        // cancelled
        continue;
      }
      now = System.currentTimeMillis();
      if (t.expirytime > now) {
        // queue rearranged
        timerQueue.put(t);
        continue;
      }
      scheduler.makeRunnable(t);                   
    }    
    //- TimerThread::run( )
  }
  
  
  /**
    Schedules a timeout by placing it in the appropriate queue.  Wakes up 
    the timerThread so that it can calculate the new wakeup time.
  */
  public synchronized void schedule(Timeout timeout) {
    //+ TimerThread::schedule(Timeout)
    timerQueue.put(timeout);
    notifyAll();
    //- TimerThread::schedule(Timeout)
  }
  
  
  /**
    Cancels a timeout by removing it from the appropriate queue. Doesn't 
    need to wake up the timerThread, that would just be some extra work. 
  */
  public synchronized void cancel(Timeout timeout) {  
    //+ TimerThread::cancel(Timeout)
    timerQueue.take(timeout);
    timeout.expirytime = 0;
    notifyAll();
    //- TimerThread::cancel(Timeout)
  }
  
  
} // TimerThread


// +*************** UNUSED CODE ***************
// -*************** UNUSED CODE ***************
