/**
  Copyright (c) 1998 Pekka Nikander & Juha Parssinen
  All rights reserved.
  XXX   %Z%%M% version %I% date:%G%
  Module:  Buffer
  Package: utility
  File:    Buffer.java
*/
package fi.hut.tcm.jacob.utility;
/**
  Buffer
  
  Buffer implements low level methods for different kind of storages like 
  queue, pool and ordered queue. This is circular.
*/
abstract class Buffer
  extends Object {

  /**
    array contains references to items which are in the storage.
  */
  private volatile Object[ ] array;
  
  /**
    This is index to head of queue in the array. It points to first full 
    slot in array unless the array is empty.
  */
  private volatile int head = 0;
  
  /**
    This is index to tail of queue in the array. It points to first empty 
    slot in array unless the array is full.
  */
  private volatile int tail = 0;
  
  /**
    The default size of array. Used in the default constructor.
  */
  private static int DEFAULT_SIZE = 8;
  
  /**
    Default constructor of the Buffer. Uses default size when create new 
    array.
  */
  protected Buffer() {
  
    //+ Buffer::Buffer( )
    this(DEFAULT_SIZE);
    //- Buffer::Buffer( )
  }
  
  
  /**
    The overloaded constructor for the Buffer. Uses given initialSize to 
    create new array.
  */
  protected Buffer(int initialSize) {
  
    //+ Buffer::Buffer(int)
    if (initialSize <= 0) {
      throw new IllegalArgumentException("Negative or zero initial buffer size");
    }
    array = new Object[initialSize];
    //- Buffer::Buffer(int)
  }
  
  
  /**
    Returns true if array is full otherwise false. This method is used 
    unsynchronized. 
  */
  protected boolean full() {
  
    //+ Buffer::full( )
    return array[tail] != null;
    //- Buffer::full( )
  }
  
  
  /**
    Return true if array is empty otherwise false. This method is used 
    unsynchronized. 
  */
  protected boolean empty() {
  
    //+ Buffer::empty( )
    return array[head] == null;
    //- Buffer::empty( )
  }
  
  
  /**
    Looks up if the item is in storage.
    @returns Index of the item, or -1
  */
  protected int has(Object item) {
  
    //+ Buffer::has(Object)
    final int length = getLength();
    for (int i = 0; i < length; i++)
      if (array[(head + i) % array.length] == item)
        return i;
    return -1;
    //- Buffer::has(Object)
  }
  
  
  /**
    Adds given object before head and moves head to point it. If array is 
    full returns false otherwise true. This should be used synchronized.
  */
  protected boolean putHead(Object item) {
  
    //+ Buffer::putHead(Object)
    return put(item, 0);
    //- Buffer::putHead(Object)
  }
  
  
  /**
    Returns object in head and removes it from the array. If the array is 
    empty return null. This should be used synchronized.
  */
  protected Object takeHead() {
  
    //+ Buffer::takeHead( )
    return take(0);
    //- Buffer::takeHead( )
  }
  
  
  /**
    Returns reference to object in head. If the array is empty returns null
    . This should be used synchronized.
  */
  protected Object getHead() {
  
    //+ Buffer::getHead( )
    return get(0);
    //- Buffer::getHead( )
  }
  
  
  /**
    Adds given object to tail and advances tail. If array is full returns 
    false otherwise true.This should be used synchronized. 
  */
  protected boolean putTail(Object item) {
  
    //+ Buffer::putTail(Object)
    return put(item, getLength());
    //- Buffer::putTail(Object)
  }
  
  
  /**
    Returns object before tail and removes it from the array. If the array 
    is empty return null.This should be used synchronized.
  */
  protected Object takeTail() {
  
    //+ Buffer::takeTail( )
    return take(getLength() - 1);
    //- Buffer::takeTail( )
  }
  
  
  /**
    Returns reference to object before tail. If the array is empty return 
    null.This should be used synchronized.
  */
  protected Object getTail() {
  
    //+ Buffer::getTail( )
    return get(getLength() - 1);
    //- Buffer::getTail( )
  }
  
  
  /**
    Puts item to (tail + offset) modulo array lenght. This should be used 
    synchronized.
  */
  protected boolean put(Object item, int offset) {
  
    //+ Buffer::put(Object, int)
    if (full()) return false;

    if (offset < 0 || offset > getLength())
      return false;

    final int index = (head + offset) % array.length;

    if (index == tail) {
      // This block is in this order because full/empty are not syncronized
      tail = (tail + 1) % array.length;
      array[index] = item;
      return true;
    }

    if (index == head) {
      // This block is in this order because full/empty are not syncronized
      final int nextHead = (head - 1 + array.length) % array.length;
      array[nextHead] = item;
      head = nextHead;
      return true;
    }

    if (index < tail) {
      // Move the rest of array one step right; Order important
      System.arraycopy(array, index, 
                       array, index + 1, 
                       tail - index);
      tail = (tail + 1) % array.length;
      array[index] = item;
      return true;
    }
    if (index > tail) {
      // Move the beginning of array one step left; Order important
      System.arraycopy(array, head, 
                       array, head - 1, 
                       index - head + 1);
      head = head - 1;
      array[index] = item;
      return true;
    }

    throw new InternalError("internal synchronization failure");
    //- Buffer::put(Object, int)
  }
  
  
  /**
    Removes the given Object from the Storage.
    @return The object removed, or null if not found.
  */
  protected Object take(Object item) {
  
    //+ Buffer::take(Object)
    final int offset = has(item);
    if (offset < 0)
      return null;

    return take(offset);
    //- Buffer::take(Object)
  }
  
  
  /**
    Removes item  from (tail + offset) modulo array lenght.This should be 
    used synchronized.
  */
  protected Object take(int offset) {
  
    //+ Buffer::take(int)
    if (offset < 0 || offset >= getLength())// Also true on empty
      return null;

    final int index = (head + offset) % array.length;
    Object returnValue = array[index];

    if (index == head) {
      // Order important
      head = (head + 1) % array.length;
      array[index] = null;
      return returnValue;
    }

    if (index < tail) {
      // Move the rest of array left; may be NOP if index == tail - 1
      System.arraycopy(array, index + 1,
                       array, index, 
                       tail - index - 1);
      {
        // Order important
        final int nextTail = (tail - 1 + array.length) % array.length;
        array[nextTail] = null;
        tail = nextTail;
      }
      return returnValue;
    }

    if (index > tail) {
      // Move the beginning of array right
      System.arraycopy(array, head, 
                       array, head + 1, 
                       index - head);
      {
        // Order important
        final int prevHead = head;
        head = (head + 1) % array.length;
        array[prevHead] = null;
      }
      return returnValue;
    }

    throw new InternalError("internal synchronization failuer");
    //- Buffer::take(int)
  }
  
  
  /**
    Returns reference to the array's item which place is (tail + offset) 
    modulo array's size. If there aren't  one returns null.  This should be
     used synchronized.
  */
  protected Object get(int offset) {
  
    //+ Buffer::get(int)
    if (offset < 0 || offset > getLength()) // XXX >=
      throw new IllegalArgumentException("XXX");
    return array[(head + offset) % array.length];
    //- Buffer::get(int)
  }
  
  
  /**
    Returns number of used slots in the array. This should be used 
    synchronized.
  */
  protected int getLength() {
  
    //+ Buffer::getLength( )
    if (full()) 
      return array.length;
    return (tail + array.length - head) % array.length;
    //- Buffer::getLength( )
  }
  
  
  /**
    Enlarges the array to be double size. This should be used synchronized.
  */
  protected boolean enlarge() {
  
    //+ Buffer::enlarge( )
    if (!full() || (array.length * 2) > Integer.MAX_VALUE)
      return false;

    Object tmpArray[] = new Object[array.length * 2];
    if (head == 0) {

      // special case when head and tail both are 0
      // head sits still and tail moves to the first empty place
      System.arraycopy(array, 0, 
                       tmpArray, 0, 
                       array.length);
      {
        // This block is in this order to avoid 
        // IndexoutOfBoudsException in empty/full
        int tmp = array.length;
        array = tmpArray; 
        tail = tmp; // full() changes value to false
      }
      return true;

    } else {

      // First we copy items from start to tail (excluding tail) ...
      System.arraycopy(array, 0, 
                       tmpArray, 0, 
                       tail);
      // ... then items from head to end (including head)
      System.arraycopy(array, head, 
                       tmpArray, tmpArray.length - (array.length - head), 
                       array.length - head);
      {
        // This block is in this order to avoid 
        // IndexoutOfBoudsException and prevent empty to return true
        // after changing used array
        final int tmpHead = head;
        final int originalSize = array.length;
        tmpArray[head] = this; // without this empty returns true
        array = tmpArray;
        head = array.length - (originalSize - tmpHead); 
        array[tmpHead] = null; // remove previous "this"
      }
      return true;

    }
    //- Buffer::enlarge( )
  }
  
  
  /**
    Debug string, prints out array's contents and size.
  */
  public String toString() {
  
    //+ Buffer::toString( )
    StringBuffer tmp = new StringBuffer();
    int i;
    for(i = 0; i < array.length; i++) {
      tmp.append("[");
      tmp.append(array[i]);
      if (i == head) 
        tmp.append("] IS HEAD\n");
      else if (i == tail)
        tmp.append("] IS TAIL\n");
      else
        tmp.append("]\n");
    }
    tmp.append("<#" + i + ">");
    return tmp.toString();               
    //- Buffer::toString( )
  }
  
  
} // Buffer


