/**
  Copyright (c) 1998 Pekka Nikander & Juha Parssinen
  All rights reserved.
  XXX   %Z%%M% version %I% date:%G%
  Module:  OrderedQueue
  Package: utility
  File:    OrderedQueue.java
*/
package fi.hut.tcm.jacob.utility;
/**
  OrderedQueue
  
  A OrderedQueue is used to store Items which have priorities (i.e. Timeout
  ).
*/
public final class OrderedQueue
  extends Queue
  implements OrderedStorage, Storage {

  /**
    Returns true if descending order otherwise false.
  */
  private boolean descending;
  
  /**
  */
  private OrderQuery query;
  
  /**
    This is overloaded constructor of theOrderedQueue. It uses defaultsize 
    for new storage and true for descending.
  */
  public OrderedQueue(OrderQuery query) {
  
    //+ OrderedQueue::OrderedQueue(OrderQuery)
    super();
    this.query = query;
    this.descending = true;
    //- OrderedQueue::OrderedQueue(OrderQuery)
  }
  
  
  /**
    This is overloaded constructor of the OrderedQueue. It uses given 
    initialSize for new storage and true for decending.
  */
  public OrderedQueue(OrderQuery query, int initialSize) {
  
    //+ OrderedQueue::OrderedQueue(OrderQuery, int)
    super(initialSize);
    this.query = query;
    this.descending = true;
    //- OrderedQueue::OrderedQueue(OrderQuery, int)
  }
  
  
  /**
    This is overloaded constructor of theOrderedQueue. It uses defaultsize 
    for new storage.
  */
  public OrderedQueue(OrderQuery query, boolean descending) {
  
    //+ OrderedQueue::OrderedQueue(OrderQuery, boolean)
    super();
    this.query = query;
    this.descending = descending;
    //- OrderedQueue::OrderedQueue(OrderQuery, boolean)
  }
  
  
  /**
    This is overloaded constructor of the OrderedQueue. It uses given 
    initialSize for new storage.
  */
  public OrderedQueue(OrderQuery query, int initialSize, boolean descending) {
  
    //+ OrderedQueue::OrderedQueue(OrderQuery, int, boolean)
    super(initialSize);
    this.query = query;
    this.descending = descending;
    //- OrderedQueue::OrderedQueue(OrderQuery, int, boolean)
  }
  
  
  /**
    Returns true if storage is full otherwise false. This method is used 
    unsynchronized
  */
  public boolean full() {
  
    //+ OrderedQueue::full( )
    return super.full();
    //- OrderedQueue::full( )
  }
  
  
  /**
    Return true if storage is empty otherwise false.  This method is used 
    unsynchronized. 
  */
  public boolean empty() {
  
    //+ OrderedQueue::empty( )
    return super.empty();
    //- OrderedQueue::empty( )
  }
  
  
  /**
    Adds item to OrderedQueue using priority and notifyes all waiting take 
    method calls. This is used synchronized.
  */
  public synchronized void put(Object item) {
  
    //+ OrderedQueue::put(Object)
    final long key = query.getOrderKey(item);

    if (full())
      enlarge();
      
    // Brute force ...
    final int length = super.getLength();
    int i;
    for (i = 0; i < length; i++) {
      final long diff = query.getOrderKey(get(i)) - key;
      if (descending && diff <= 0 || !descending && diff >= 0) {
        super.put(item, i);
        notifyAll();
        // System.out.println("#Buffer :\n" + super.toString()); // XXX
        return;
      }
    }
    // i == length
    putTail(item);
    notifyAll();
    // System.out.println("#Buffer :\n" + super.toString()); // XXX
    return; 
    //- OrderedQueue::put(Object)
  }
  
  
  /**
    Removes one Item (highest priority) from OrderedQueue and returns 
    reference to it. If the OrderedQueue is empty this method call blocks 
    until notifyed (by put method). This is used synchronized.
  */
  public synchronized Object take() {
  
    //+ OrderedQueue::take( )
    return super.take(); // This make notifyAll();
    //- OrderedQueue::take( )
  }
  
  
  /**
    Removes given object from storage and returns reference to it. If item 
    doesn't found returns null.
  */
  public synchronized Object take(Object item) {
  
    //+ OrderedQueue::take(Object)
    Object returnItem = super.take(item);
    if (returnItem != null)
      notifyAll();
    return returnItem;
    //- OrderedQueue::take(Object)
  }
  
  
  /**
    Returns reference to item which have highest key (i.e. priority). 
    Returns null if storage is empty.
  */
  public synchronized Object get() {
  
    //+ OrderedQueue::get( )
    return super.getHead();
    //- OrderedQueue::get( )
  }
  
  
} // OrderedQueue


// +*************** UNUSED CODE ***************
// -*************** UNUSED CODE ***************
